/*
 * Implementation of a plugin module to
 * access the ArtNet protocol.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>

#include <errno.h>

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/ioctl.h>
#include <linux/if.h>

#include <dmx/plugin.h>


#include "Art-Net.h" /* Header-File from Artistic Licence */



/*
 * ---------[ Prototypes ]----------
 */
static int Initialize (char * hostname);

static void printOutputRequest (T_ArtDmx *data);
static void printPollRequest   (T_ArtPoll *poll);
static void printPollReply     (T_ArtPollReply *reply);
static void *runArtNetThread   (void *arg);

static unsigned short H2INS(unsigned short x);
static int isMyself (struct sockaddr_in *addr);

void add_interfaces2nonhosts();

static int    sendDMX         (int universe, unsigned char *dmxbuffer, int dmxlen,
                               struct sockaddr *sa, socklen_t len);
static void   broadcastDMX    (int universe, unsigned char *dmxbuffer, int dmxlen);
static void   sendPollRequest (struct sockaddr *sa, socklen_t len);
static void   sendPollReply   (struct sockaddr *sa, socklen_t len);
static int    startThread     ();
static int    stopThread      ();
static char * IP2str          (char *str, struct sockaddr_in *addr);




/*
 * -------------[ plugin wide variables ]-------------
 */

static  dmxuniverse_t       *universes[15*4];
static  int max_universes =  sizeof(universes) / sizeof(dmxuniverse_t *);

static  pthread_t rxtx_thread;
static  char      thread_active = 0;


static  int subnet = 15;
static  int debug  = 0;
static  int sock   = -1;


static  struct sockaddr_in servaddr;
static  struct sockaddr_in anyaddr;
static  struct sockaddr_in broadcast;


/*
 * list of socketaddresses that should be ignored
 */
static  struct sockaddr_in  nohosts[10]; // buggy constant, but will work now.
static  int    num_nohosts = 0;




/*
 * -------------[ plugin methods ]-------------
 */

/*
 * This method is called by the plugin interface after the
 * module has been loaded.
 */
int initialize (void)
{
  int i;

  for (i=0; i<max_universes; i++)
    universes[i] = NULL;

  if (!Initialize (NULL))
    return 0;
  return -1;
}

/*
 * This method is called by the plugin interface before the
 * module will be unloaded.
 */
int finalize ()
{
  int i;

  if (thread_active)
    {
      stopThread ();
      thread_active = 0;
    }


  for (i=0; i<max_universes; i++)
    if (universes[i])
      universes[i] = NULL;

  if (sock != -1)
    {
      shutdown (sock, 3);
      close (sock);
    }
}


/*
 * This method binds a universe object to
 * the universe number <nr>.
 * Returns 0 on success or -1 on error.
 */
int bind_universe (int nr, dmxuniverse_t *u)
{
  if (nr>=0 && nr<max_universes && !universes[nr])
    {
      if (!thread_active)
        startThread ();
      universes[nr] = u;
      return 0;
    }
  return -1;
}


/*
 * This method tries release the binding between
 * a universe object and the universe number <nr>.
 * Returns 0 on success or -1 on error.
 */
int unbind_universe (int nr, dmxuniverse_t *u)
{
  if (nr>=0 && nr<max_universes && universes[nr] && universes[nr]==u)
    {
      universes[nr] = NULL;
      return 0;
    }
  return -1;
}


/*
 * Sends out the universe data of the universe
 * object <u> on the universe <nr>.
 */
int send_universe (int nr, dmxuniverse_t * u)
{
  if (nr>=0 && nr<max_universes /* && universes[nr] && universes[nr]==u */)
    return sendDMX (nr, u->buffer, u->numslots, (struct sockaddr *)&broadcast, sizeof(broadcast));
  return -1;
}


/*
 * Get the some information about the universe.
 */
int get_plugin_info (dmxplugin_info_t *info)
{
  if (info)
    {
      strcpy (info->vendor_name, "ArtisticLicence Ltd.");
      strcpy (info->proto_name, "ArtNet");
      info->proto_release = 1.0;

      info->release = 1.0;
      strcpy (info->author, "Michael Stickel");
      strcpy (info->email,  "michael@cubic.org");
      info->num_universes = max_universes;
      return 0;
    }
  return -1;
}

#if 0
dmxuniverse_info_t * get_universe_info ()
{
  return NULL;
}
#endif

/*-------------------------[ArtNetTranceiver.cc]--------------------
 */


/*
 * Host 2 Inverse Network byte order
 * This is the only way to make it working on other machines, like SPARC.
 */
static unsigned short H2INS(unsigned short x)
{
  unsigned short y = htons(x);
  return ((y&0xff)<<8) | ((y>>8)&0xff);
}


static int getInetAddrByNameAndPort (struct sockaddr_in *sa, size_t slen, unsigned short port, char *name)
{
  if (sa && slen==sizeof(struct sockaddr_in))
    {
      struct hostent * he = gethostbyname(name);
      if (he)
        {
          bzero (sa, slen);
          sa->sin_family = AF_INET;
          sa->sin_port = port;
          memcpy (&sa->sin_addr.s_addr, he->h_addr, he->h_length);
          return 0;
        }
    }
  return -1;
}


static int Initialize (char * hostname)
{
  int  one    = 1;
  int *oneptr = &one;


  sock = socket (AF_INET, SOCK_DGRAM, 0);
  if (sock<0)
    {
      printf ("Error createing socket: %s\n", strerror(errno));
      return -1;
    }
  
  if (setsockopt (sock, SOL_SOCKET, SO_BROADCAST, oneptr, sizeof(one)) == -1)
    {
      fprintf (stderr, "Error getting broadcast permission: %s\n", strerror(errno));
    }

  getInetAddrByNameAndPort (&servaddr,  sizeof(servaddr),  htons(DefaultPort), hostname?hostname:"255.255.255.255");
  getInetAddrByNameAndPort (&anyaddr,   sizeof(anyaddr),   htons(DefaultPort), "0.0.0.0");
  getInetAddrByNameAndPort (&broadcast, sizeof(broadcast), htons(DefaultPort), "255.255.255.255");


  memcpy (&nohosts[num_nohosts++], &anyaddr, sizeof(anyaddr));
  memcpy (&nohosts[num_nohosts++], &broadcast, sizeof(broadcast));

  add_interfaces2nonhosts ();

  return 0;
}

/*
 * add the addresses of all network interfaces into the list of ip's to ignore
 */
void add_interfaces2nonhosts()
{
  struct ifconf ifc;
  struct ifreq  ifr[50];

  int sockfd = socket(AF_INET, SOCK_DGRAM,0);
  if (sockfd!=-1)
    {
      ifc.ifc_len = sizeof(ifr);
      ifc.ifc_req = ifr;
      if (ioctl (sockfd, SIOCGIFCONF, &ifc) >= 0)
        {
          int count = ifc.ifc_len / sizeof(struct ifreq);
          int i;

          for (i=0; i<count; i++)
            {
              memcpy (&nohosts[num_nohosts], &ifr[i].ifr_addr, sizeof(struct sockaddr_in));
              nohosts[num_nohosts].sin_port = htons(DefaultPort);
              nohosts[num_nohosts].sin_family = AF_INET;

              num_nohosts++;
            }
        }
      close(sockfd);
    }
}





static int isMyself (struct sockaddr_in *addr)
{
  if (addr)
    {
      int i;
      for (i=0; i<num_nohosts; i++)
        if (addr->sin_addr.s_addr==nohosts[i].sin_addr.s_addr && addr->sin_port==nohosts[i].sin_port)
          return 1;
      return 0;
    }
  return -1;
}





char *IP2str (char *str, struct sockaddr_in *addr)
{
  if (str)
    {
      if (addr)
	{
	  sprintf (str, "%s:%d", inet_ntoa (addr->sin_addr), ntohs(addr->sin_port));
	}
      else
	strcpy (str, "unknown");
    }
  return str;
}



static int startThread ()
{
  if (!thread_active)
    {
      int stat;
      static int bound=0;

      if (!bound)
        {
          if (bind (sock, (struct sockaddr *)&anyaddr, sizeof(anyaddr)) < 0)
            {
              int err = errno;
              char * str = "unknown";
              if (err==EINVAL)
	        str = "EINVAL";
              else if (err==EACCES)
	        str = "EACCES";
              printf ("Error while bind %s\n", str);
            }
          else
            bound=1;
        }

      stat = pthread_create (&rxtx_thread, NULL, runArtNetThread, NULL);
      if (stat)
	{
	  thread_active = 0;
	  printf ("unable to create thread\n");
	}
      else
	{
	  thread_active = 1;
	  return 0;
	}
    }
  return -1;
}


static int stopThread ()
{
  if (thread_active)
    {
      int stat = pthread_kill (rxtx_thread, SIGINT);
      if (!stat || stat == ESRCH)
	{
	  void *ret = NULL;
	  if (!pthread_join (rxtx_thread, &ret))
	    {
	      printf ("thread returned successfully and returned value %p\n", ret);
	      return 0;
	    }
	}
    }
  return -1;
}


static void *runArtNetThread (void *arg)
{
  int  n;
  unsigned char buffer[2000];
  
  while (1)
    {
      struct sockaddr_in claddr;
      socklen_t          cllen = sizeof(claddr);


      if ((n = recvfrom (sock, buffer, sizeof(buffer), 0, (struct sockaddr *)&claddr, &cllen)) < 0)
	{
	  printf ("Error on receive: %s\n", strerror(errno));
	}


      if (isMyself (&claddr) > 0)
        continue;


      if (debug)
	{
	  printf ("got message\n");
	}
 
      switch (((T_ArtPoll *)buffer)->OpCode)
	{
	case OpPollReply:
	  if (n >= sizeof(T_ArtPollReply))
	    {
	      if (debug)
		{
		  printPollReply ((T_ArtPollReply *)buffer);
		}
	    }
	  break;
	      
	case OpPoll:
	  if (n >= sizeof(T_ArtPoll))
	    {
	      T_ArtPoll *poll = (T_ArtPoll *)buffer;

	      if (debug)
		{
		  printPollRequest (poll);
		}

	      if ((poll->TalkToMe&1) || cllen <= 0)
		{
		  sendPollReply ((struct sockaddr *)&broadcast, sizeof(broadcast));
		}
	      else
		{
		  if (debug && cllen == sizeof(struct sockaddr_in))
		    {
		      char str[20];
		      IP2str (str, &claddr);
		      printf ("reply to %s\n", str);
		    }
		  sendPollReply ((struct sockaddr *)&claddr, cllen);
		}
	    }
	  break;


	case OpOutput:
	  if (n >= sizeof(T_ArtDmx) /* && claddr.sin_port != broadcast.sin_port*/ )  // ignore data coming from this port
	    {
	      T_ArtDmx *data = (T_ArtDmx *)buffer;
	      
	      if (memcmp (data->ID, "Art-Net", 7) == 0 &&
		  ((data->Universe>>4)&0xf) == subnet)
		{
		  size_t length = ntohs(data->Length);
		  if (debug)
		    printOutputRequest (data);

                  if (universes[data->Universe&3])
                    {
                      dmxuniverse_t *u = universes[data->Universe&3];

                      u->numslots = length;
                      memcpy (u->buffer, data->Data, length);

                      if (universes[data->Universe&3]->notify)
                        universes[data->Universe&3]->notify(universes[data->Universe&3]);
                    }
		}
	    }
	  break;


	default:
	  break;
	  
	}
    }
}



int sendDMX (int universe, unsigned char *dmxbuffer, int dmxlen,
				struct sockaddr *sa, socklen_t len)
{
  T_ArtDmx data;

  struct sockaddr_in tsa;

  strcpy ((char *)data.ID, "Art-Net");
  data.OpCode = H2INS(OpOutput);
  data.VersionH = 0;
  data.Version  = 14;
  data.Filler1  = 0;
  data.Physical = 0;
  data.Universe = subnet<<4 | universe;
  data.Length   = htons (dmxlen);
  memcpy (data.Data, dmxbuffer, dmxlen);

  if (debug)
    {
      printf ("sendDMX\n");
    }

  getInetAddrByNameAndPort (&tsa, sizeof(tsa), htons(DefaultPort), "195.126.101.63");

  
  if (sendto (sock, &data, sizeof(data), 0, (struct sockaddr *)&tsa, sizeof(tsa)) < 0)
    {
      printf ("ArtNetTranceiver::sendDMX : Error sending: %s\n", strerror(errno));
      return -1;
    }
  return 0;
}


/*
 * Poll for an ArtNet device
 */
void sendPollRequest (struct sockaddr *sa, socklen_t len)
{
  T_ArtPoll poll;

  strcpy ((char *)poll.ID, "Art-Net");
  poll.OpCode = H2INS(OpPoll);
  poll.VersionH = 0;
  poll.Version = 14;
  poll.TalkToMe = 2;
  poll.Pad = 0;

  if (debug)
    {
      printf ("sendPollRequest\n");
    }
      
  if (sendto (sock, &poll, sizeof(poll), 0, sa, len) < 0)
    {
      printf ("sendPollRequest : Error sending: %s\n", strerror(errno));
    }
}


/*
 * Reply to an ARtNet Poll.
 */
void sendPollReply (struct sockaddr *sa, socklen_t len)
{
  int  i;
  T_ArtPollReply reply;
  
  memcpy (reply.ID, "Art-Net", 8);
  reply.OpCode = H2INS(OpPollReply);

  reply.BoxAddr.IP[0] = 2;
  reply.BoxAddr.IP[1] = 0;
  reply.BoxAddr.IP[2] = 0;
  reply.BoxAddr.IP[3] = 99;
  reply.BoxAddr.Port  = DefaultPort;

  reply.VersionInfoH = 0;
  reply.VersionInfo  = 5;

  reply.SubSwitchH = 0;
  reply.SubSwitch = subnet;
    
  reply.OemH = 0;
  reply.Oem = 0;
    
  reply.EstaManH = 0;
  reply.EstaMan = 0;

  memcpy(reply.ShortName,  "LLGArtNetDMXNode\0", 18);
  memcpy(reply.LongName,   "LLG ArtNet DMX Node for Linux (http://llg.cubic.org/llgann/)   ", 64);
  memcpy(reply.NodeReport, "Hello master. I am feeling perfect and waiting for your advice ", 64);

  reply.NumPortsH = 0;
  reply.NumPorts  = 4;
  for (i=0; i<4; i++)
    {
      reply.PortTypes[i] = 0xc0;
      reply.GoodInput[i] = 0x80;
      reply.GoodOutput[i] = 0x80;
	reply.Swin[i] = 0xF4+i;
	reply.Swout[i] = 0xF0+i;
    }
  reply.SwVideo = 0;
  reply.SwMacro = 0x55;
  reply.SwRemote = 0xAA;

  if (debug)
    {
      printf ("sendPollReply\n");
    }

  if (sendto (sock, &reply, sizeof(reply), 0, sa, len) < 0)
    {
      printf ("sendPollReply : Error sending: %s\n", strerror(errno));
    }
}


/*
 * ------------[ dumping methods (local) ]----------------
 */

static void printOutputRequest (T_ArtDmx *data)
{
  if (debug)
    {
      int i;
      int length = ntohs(data->Length);
      
      printf ("DMX-Data\n");
      printf ("Version:  %d.%d\n", data->VersionH, data->Version);
      printf ("Filler1:  %d\n", data->Filler1);
      printf ("Physical: %d\n", data->Physical);
      printf ("Universe: subnet=%d, wheel=%d\n", (data->Universe>>4)&0xf, data->Universe&0xf);
      printf ("Length:   %d\n", length);
      printf ("Data:     \n");
      for (i=0; i<20 && i<MaxDataLength; i++)
	printf ("%02X ", data->Data[i]);
    }
}


static void printPollRequest (T_ArtPoll *poll)
{
  printf ("ArtPoll\n");
  printf ("ID: %s\n", poll->ID);
  printf ("Version: %d.%d\n", poll->VersionH, poll->Version);
  printf ("TalkToMe: %s %s\n", 
	  (poll->TalkToMe&1)?"reply to server IP":"reply broadcast",
	  (poll->TalkToMe&2)?"sends reply when it needs to":"reply when polled");
}


/*
 * Display the values of an ArtNet Poll reply.
 */
static void printPollReply (T_ArtPollReply *reply)
{
  int i;

  printf ("ArtPollReply\n");

  printf ("Kennung: %s\n", reply->ID);
  printf ("BoxAddr: %d.%d.%d.%d:%d\n", reply->BoxAddr.IP[0], reply->BoxAddr.IP[1], reply->BoxAddr.IP[2], reply->BoxAddr.IP[3], reply->BoxAddr.Port);
  printf ("VersionInfo: %d.%d\n", reply->VersionInfoH, reply->VersionInfo);
  printf ("SubSwitch: %d.%d\n", reply->SubSwitchH, reply->SubSwitch);
  printf ("Oem: %d.%d\n", reply->OemH, reply->Oem);
  printf ("EstaMan: %d.%d\n", reply->EstaManH, reply->EstaMan);
    
  printf ("ShortName: %s\n", reply->ShortName);
  printf ("LongName: %s\n", reply->LongName);
  printf ("Report: %s\n", reply->NodeReport);
  
  printf ("NumPorts = %d.%di:\n", reply->NumPortsH, reply->NumPorts);
  for (i=0; i<reply->NumPorts; i++)
    {
      printf ("type=%02X,in=%02X,out=%02X,Swin=%02X,Swout=%02X\n", reply->PortTypes[i],
	      reply->GoodInput[i], reply->GoodOutput[i], reply->Swin[i], reply->Swout[i]);
    }
    
  printf ("SwVideo=%02X\n", reply->SwVideo);
  printf ("SwMacro=%02X\n", reply->SwMacro);
  printf ("SwRemote=%02X\n", reply->SwRemote);
}


