/*
 * Implementation of a plugin module to
 * access the AVAB/UDP 1.6 protocol.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>

#include <errno.h>

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/ioctl.h>
#include <linux/if.h>

#include <dmx/plugin.h>


#include "avabudp.h"


/*
 * ---------[ Prototypes ]----------
 */
static int Initialize (char * hostname);

static void *AvabUDP_Thread (void *arg);

static int isMyself (struct sockaddr_in *addr);

void add_interfaces2nonhosts();

static int    sendDMX         (int universe, unsigned char *dmxbuffer, int dmxlen,
                               struct sockaddr *sa, socklen_t len);

static int    activateReceiver ();
static int    stopReceiver     ();

static char * IP2str          (char *str, struct sockaddr_in *addr);




/*
 * -------------[ plugin wide variables ]-------------
 */

static  dmxuniverse_t       *universes[AVAB_UDP_N_SUBNET * AVAB_UDP_N_UNIV];
static  const int max_universes =  sizeof(universes) / sizeof(dmxuniverse_t *);
static  int  used_universes = 0;

static  pthread_t rxtx_thread;
static  char      thread_active = 0;


static  int subnet = 15;
static  int debug  = 0;
static  int sock   = -1;


static  struct sockaddr_in servaddr;
static  struct sockaddr_in anyaddr;
static  struct sockaddr_in broadcast;


/*
 * list of socketaddresses that should be ignored
 */
static  struct sockaddr_in  nohosts[10]; // buggy constant, but will work now.
static  int    num_nohosts = 0;

static  seqnum_t sequences[AVAB_UDP_N_UNIV];




/*
 * -------------[ plugin methods ]-------------
 */

/*
 * This method is called by the plugin interface after the
 * module has been loaded.
 */
int initialize (void)
{
  int i;

  for (i=0; i<max_universes; i++)
    universes[i] = NULL;

  memset (sequences, 0, sizeof(sequences));

  if (!Initialize (NULL))
    return 0;
  return -1;
}

/*
 * This method is called by the plugin interface before the
 * module will be unloaded.
 */
int finalize ()
{
  int i;

  if (thread_active)
    {
      stopReceiver ();
    }

  for (i=0; i<max_universes; i++)
    if (universes[i])
      universes[i] = NULL;

  if (sock != -1)
    {
      shutdown (sock, 3);
      close (sock);
    }
}


/*
 * This method binds a universe object to
 * the universe number <nr>.
 * Returns 0 on success or -1 on error.
 */
int bind_universe (int nr, dmxuniverse_t *u)
{
  if (nr>=0 && nr<max_universes && !universes[nr])
    {
      if (used_universes==0)
        {
          activateReceiver();
        }
      used_universes++;

      universes[nr] = u;
      return 0;
    }
  return -1;
}


/*
 * This method tries release the binding between
 * a universe object and the universe number <nr>.
 * Returns 0 on success or -1 on error.
 */
int unbind_universe (int nr, dmxuniverse_t *u)
{
  if (nr>=0 && nr<max_universes && universes[nr] && universes[nr]==u)
    {
      universes[nr] = NULL;
      used_universes--;
      if (used_universes<=0)
        {
          /* should stop the thread */
        }
      return 0;
    }
  return -1;
}


/*
 * Sends out the universe data of the universe
 * object <u> on the universe <nr>.
 */
int send_universe (int nr, dmxuniverse_t * u)
{

  if (nr>=0 && nr<max_universes /* && universes[nr] && universes[nr]==u */)
    {
      return sendDMX (nr, u->buffer, u->numslots, (struct sockaddr *)&broadcast, sizeof(broadcast));
    }
  return -1;
}


/*
 * Get the some information about the universe.
 */
int get_plugin_info (dmxplugin_info_t *info)
{
  if (info)
    {
      strcpy (info->vendor_name, "Transtechnik GmbH");
      strcpy (info->proto_name, "AVAB/UDP");
      info->proto_release = 1.6;

      info->release = 1.0;
      strcpy (info->author, "Michael Stickel");
      strcpy (info->email,  "michael@cubic.org");
      info->num_universes = max_universes;
      return 0;
    }
  return -1;
}



/*
 * This should go into some inet-tools library.
 */
static int getInetAddrByNameAndPort (struct sockaddr_in *sa, size_t slen, unsigned short port, char *name)
{
  if (sa && slen==sizeof(struct sockaddr_in))
    {
      struct hostent * he = gethostbyname(name);
      if (he)
        {
          bzero (sa, slen);
          sa->sin_family = AF_INET;
          sa->sin_port = port;
          memcpy (&sa->sin_addr.s_addr, he->h_addr, he->h_length);
          return 0;
        }
    }
  return -1;
}


static int Initialize (char * hostname)
{
  int  one    = 1;
  int *oneptr = &one;


  sock = socket (AF_INET, SOCK_DGRAM, 0);
  if (sock<0)
    {
      printf ("Error createing socket: %s\n", strerror(errno));
      return -1;
    }
  
  if (setsockopt (sock, SOL_SOCKET, SO_BROADCAST, oneptr, sizeof(one)) == -1)
    {
      fprintf (stderr, "Error getting broadcast permission: %s\n", strerror(errno));
    }

  getInetAddrByNameAndPort (&servaddr,  sizeof(servaddr),  htons(AVAB_UDP_DEFAULT_PORT), hostname?hostname:"255.255.255.255");
  getInetAddrByNameAndPort (&anyaddr,   sizeof(anyaddr),   htons(AVAB_UDP_DEFAULT_PORT), "0.0.0.0");
  getInetAddrByNameAndPort (&broadcast, sizeof(broadcast), htons(AVAB_UDP_DEFAULT_PORT), "255.255.255.255");

  memcpy (&nohosts[num_nohosts++], &anyaddr, sizeof(anyaddr));
  memcpy (&nohosts[num_nohosts++], &broadcast, sizeof(broadcast));

  add_interfaces2nonhosts ();

  return 0;
}

/*
 * add the addresses of all network interfaces into the list of ip's to ignore
 */
void add_interfaces2nonhosts()
{
  struct ifconf ifc;
  struct ifreq  ifr[50];

  int sockfd = socket(AF_INET, SOCK_DGRAM,0);
  if (sockfd!=-1)
    {
      ifc.ifc_len = sizeof(ifr);
      ifc.ifc_req = ifr;
      if (ioctl (sockfd, SIOCGIFCONF, &ifc) >= 0)
        {
          int count = ifc.ifc_len / sizeof(struct ifreq);
          int i;

          for (i=0; i<count; i++)
            {
              memcpy (&nohosts[num_nohosts], &ifr[i].ifr_addr, sizeof(struct sockaddr_in));
              nohosts[num_nohosts].sin_port = htons(AVAB_UDP_DEFAULT_PORT);
              nohosts[num_nohosts].sin_family = AF_INET;

              num_nohosts++;
            }
        }
      close(sockfd);
    }
}





static int isMyself (struct sockaddr_in *addr)
{
  if (addr)
    {
      int i;

      for (i=0; i<num_nohosts; i++)
        if (addr->sin_addr.s_addr==nohosts[i].sin_addr.s_addr && addr->sin_port==nohosts[i].sin_port)
          return 1;
      return 0;
    }
  return -1;
}





char *IP2str (char *str, struct sockaddr_in *addr)
{
  if (str)
    {
      if (addr)
	{
	  sprintf (str, "%s:%d", inet_ntoa (addr->sin_addr), ntohs(addr->sin_port));
	}
      else
	strcpy (str, "unknown");
    }
  return str;
}



static int activateReceiver ()
{
  if (!thread_active)
    {
      int stat=0;

      if (bind (sock, (struct sockaddr *)&anyaddr, sizeof(anyaddr)) < 0)
        {
          int err = errno;
          char * str = "unknown";
          if (err==EINVAL)
            str = "EINVAL";
          else if (err==EACCES)
            str = "EACCES";
          printf ("Error while bind %s\n", str);
          return -1;
        }

      stat = pthread_create (&rxtx_thread, NULL, AvabUDP_Thread, NULL);
      if (stat)
	{
	  thread_active = 0;
	  printf ("unable to create thread\n");
	}
      else
	{
	  thread_active = 1;
	  return 0;
	}
    }
  return -1;
}


static int stopReceiver ()
{
  if (thread_active)
    {
      int stat = pthread_kill (rxtx_thread, SIGINT);
      if (!stat || stat == ESRCH)
	{
	  void *ret = NULL;
	  if (!pthread_join (rxtx_thread, &ret))
	    {
	      printf ("thread returned successfully and returned value %p\n", ret);
              thread_active = 0;
	      return 0;
	    }
	}
    }
  return -1;
}




static int AvabUDP_ntoh (const unsigned char *buffer, const size_t len, AvabUDPDatagram_t *out)
{
  if (buffer && out && len>=10+sizeof(AvabUDPFrame_t)) /* the minimum size must include one frame */
  {
    AvabUDPDatagram_t *in = (AvabUDPDatagram_t *)buffer;

    out->magic_number   = ntohs(in->magic_number);
    out->telegram_type  = ntohs(in->telegram_type);
    out->source_address = ntohl(in->source_address);
    out->frames_count   = ntohs(in->frames_count);
    if (out->frames_count>=1 && out->frames_count<=2)
      {
	int i;

	if (out->frames_count*sizeof(AvabUDPFrame_t) + 10 > len)
	  return -1; /* UDP-packet is to short */

	for (i=0; i<out->frames_count; i++)
	  {
	    AvabUDPFrame_t *outf = &out->frames[i];
	    AvabUDPFrame_t *inf  = &in->frames[i];

	    outf->sequence = ntohs(inf->sequence);
	    outf->clock    = ntohl(inf->clock);
	    outf->unused   = ntohs(inf->unused);
	    outf->dirty_flag = ntohl(inf->dirty_flag);
	    outf->date     = ntohl(inf->date);
	    outf->frame_type = ntohs(inf->frame_type);
	    outf->universe = ntohs(inf->universe);
	    memcpy (outf->values, inf->values, 512);
	  }
	return 0;
      }
  }
  return -1;
}

/*
 * Converts an AvabUDP datagram to Network Byte Order
 * and stores it to the buffer that must be at lease 1074 bytes of size.
 */
static int AvabUDP_hton (const AvabUDPDatagram_t *in, const size_t len, unsigned char *buffer)
{
  if (buffer && in && len>=10+sizeof(AvabUDPFrame_t)) /* must be enough space to put in one dmx-frame */
  {
    AvabUDPDatagram_t *out = (AvabUDPDatagram_t *)buffer;

    out->magic_number   = htons(in->magic_number);
    out->telegram_type  = htons(in->telegram_type);
    out->source_address = htonl(in->source_address);
    out->frames_count   = htons(in->frames_count);
    if (in->frames_count>=1 && in->frames_count<=2)
      {
	int i;
	for (i=0; i<in->frames_count; i++)
	  {
	    AvabUDPFrame_t       *outf = &out->frames[i];
	    const AvabUDPFrame_t *inf  = &in->frames[i];

	    outf->sequence = htons(inf->sequence);
	    outf->clock    = htonl(inf->clock);
	    outf->unused   = htons(inf->unused);
	    outf->dirty_flag = htonl(inf->dirty_flag);
	    outf->date     = htonl(inf->date);
	    outf->frame_type = htons(inf->frame_type);
	    outf->universe = htons(inf->universe);
	    memcpy (outf->values, inf->values, 512);
	  }
	return 10 + sizeof(AvabUDPFrame_t) * in->frames_count; /* return the real size of the datagram */
      }
  }
  return -1;
}





static void *AvabUDP_Thread (void *arg)
{
  int               dglen;
  unsigned char     buffer[2000];
  AvabUDPDatagram_t data;

  /*
   * we have to setup the sceduleing policy to fifo to have
   * a quicker response.
   */

  while (1)
    {
      struct sockaddr_in claddr;
      socklen_t          cllen = sizeof(claddr);

#if 0
      /*  socks[10] */
      int n=-1;
      FD_ZERO(&rd_set);
      for (i=0; i<10; i++)
	{
	  if (socks[i]>n)
	    n=socks[i];
	  FD_SET(socks[i], &rd_set);
	}
      tv.tv_sec = 1;
      tv.tv_usec = 0;
      int stat = socket(n+1, &rd_set, NULL, NULL, &tv);
      if (stat<0)
	error ();
      else if (stat==0)
	timeout ();
      else
	{
	  for (i=0; i<10; i++)
	    {
	      if (FD_ISSET(socks[i], &rd_set))
		{
		  ComputeAvabUDPNetwork (i, socks[i]);
		}
	    }
	}
#endif
      dglen = recvfrom (sock, buffer, sizeof(buffer), 0, (struct sockaddr *)&claddr, &cllen);
      if (dglen==0)
        {
          printf ("recvfrom=0\n");
        }
      else if (dglen < 0)
	{
          int e = errno;
	  printf ("Error on receive:%d:%d: %s\n", dglen, e, strerror(e));
	}
      else
	if (dglen>=542 && !isMyself (&claddr))
	  {
	    if (debug)
	      printf ("got AvabUDP-datagram\n");
	    
	    if (AvabUDP_ntoh(buffer, dglen, &data)==0)
	      {
		if (data.magic_number==AVAB_UDP_MAGIC && data.telegram_type==AVAB_UDP_TELEGRAM_TYPE)
		  {
		    int i;
		    for (i=0; i<data.frames_count; i++)
		      {
			if (data.frames[i].frame_type==AVAB_FRAME_OUTPUT_UNPACKED &&
			    data.frames[i].universe>=AVAB_UDP_UNIV_MIN &&
			    data.frames[i].universe<=AVAB_UDP_UNIV_MAX)
			  {
			    dmxuniverse_t *u = universes[data.frames[i].universe-AVAB_UDP_UNIV_MIN];
#if 0
			    printf ("received frame with sequence 0x%04X\n", data.frames[i].sequence);
#endif
			    if (u && u->notify)
			      {
				u->numslots = 512;
				memcpy (u->buffer, data.frames[i].values, 512);
				u->notify (u);
			      }
			  }
		      }
		  }
	      }
	  }
    }
}

int sendDMX (int nr, unsigned char *dmxbuffer, int dmxlen, struct sockaddr *tsa, socklen_t len)
{
  if (dmxbuffer && dmxlen>0 && tsa && len>0)
    {
      int  net_id = nr>>5;
      int  universe = nr&0x31;
      struct sockaddr_in  sa;
      AvabUDPDatagram_t   data;

      if (AVAB_UDP_SUBNET_PORT(net_id) < 0)
        return -1;

      universe+=AVAB_UDP_UNIV_MIN;

      if (universe<AVAB_UDP_UNIV_MIN || universe>AVAB_UDP_UNIV_MAX ||
	  net_id<AVAB_UDP_SUBNET_MIN || net_id>AVAB_UDP_SUBNET_MAX)
	return -1;

      if (len > sizeof(sa))
	return -1;

      memcpy (&sa, tsa, len); /* copy template to real socket address buffer */
/*
      getInetAddrByNameAndPort (&sa, sizeof(sa), htons(AVAB_UDP_DEFAULT_PORT), "255.255.255.255");
*/
      sa.sin_port = htons(AVAB_UDP_SUBNET_PORT(net_id));
      sa.sin_family = AF_INET;

      memset (&data, 0, sizeof(data));

      data.magic_number   = htons(AVAB_UDP_MAGIC);
      data.telegram_type  = htons(AVAB_UDP_TELEGRAM_TYPE);
      data.source_address = htonl(0L); /* eigenne IP-Adresse ermitteln und eintragen */
      data.frames_count = htons(1);

      data.frames[0].sequence = htons(sequences[universe]);
      sequences[universe]++; /* we don't know how often htons uses it's parameter, so it may be incremented twice or more often */
#if 0
      data.frames[0].clock  = htonl(0L);
      data.frames[0].unused = htons(0);
      data.frames[0].dirty_flag = htonl(0L);
      data.frames[0].date = htonl(0L);
#endif
      data.frames[0].frame_type = htons(AVAB_FRAME_OUTPUT_UNPACKED);
      data.frames[0].universe   = htons(universe);
      memcpy(data.frames[0].values, dmxbuffer, (dmxlen>512)?512:dmxlen);

      if (sendto (sock, &data, sizeof(data)-sizeof(data.frames[1]), 0, (struct sockaddr *)&sa, sizeof(sa)) < 0)
	{
	  printf ("avabudp.sendDMX : Error sending: %s\n", strerror(errno));
          printf ("sock=%d\n", sock);
	}
    }
}

